# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
import time

import numpy as np
from scipy.optimize import minimize

from metku.optimization.solvers.optsolver import OptSolver


class SLSQP(OptSolver):

    def __init__(self):
        super().__init__()

    def calc_constraints(self, x=None):
        """
        Calculates constraint values and saves them to numpy array

        Parameters:
        ------------
        :param x: (Optional, default: self.X) Point where constraint values are calculated

        """
        constr_vals = []
        if x is None:
            x = self.X
        # Calculate constraints
        self.constr_vals = self.problem.eval_cons(x)

    def _create_eqcons(self):
        """ Creates a list of equality constraints

        :return: constraints
        """
        eqcons = []

        for con in self.problem.cons:
            if con.type == "=":
                eqcons.append(con)

        return eqcons

    def _create_ieqcons(self):
        """ Creates a list of inequality constraints

        :return: list of equality constraint functions
        """
        ieqcons = []

        for con in self.problem.cons:
            if con.type == "<" or con.type == "<=":
                ieqcons.append(con.neg_call)
            elif con.type == ">" or con.type == ">=":
                ieqcons.append(con)

        return ieqcons

    def _create_bounds(self):
        """ Creates a list of optimization variables' bounds

        :return: list of boundaries (lb, ub)
        """
        bounds = []
        for var in self.problem.vars:
            bound = (var.lb, var.ub)
            bounds.append(bound)

        return bounds

    def solve(self, problem, maxiter=200, x0=None, verb=False):
        """
        Solves given problem

        :param problem: problem to be solved
        :param maxiter: number of maximum iterations
        :param x0: initial guess

        :return: fopt, xopt
        """        
        
        self.verb = verb
        self.problem = problem
        self.maxiter = maxiter
        bounds = self._create_bounds()
        eqcons = self._create_eqcons()
        ieqcons = self._create_ieqcons()
        
        #print(bounds)
        
        # Create initial guess if one isn't provided
        if x0 is None:
            x0 = []
            for var in problem.vars:
                x0.append(var.ub)  # * np.random.uniform())
        problem.substitute_variables(x0)
        """
            'eps' is the step length in numerical differentiation
        """
        if verb == True:
            iprint = 2
        else:
            iprint = 1
        
        options = {'maxiter': maxiter,
                   'disp': verb,                   
                   'iprint': iprint,
                   'eps': 1e-6,
                   'ftol': 1e-5}

        constraints = []

        for eqcon in eqcons:
            con = {'type': 'eq', 'fun': eqcon}
            constraints.append(con)

        for ineqcon in ieqcons:
            con = {'type': 'ineq', 'fun': ineqcon}
            constraints.append(con)

        self.start = time.time()        
        out = minimize(self.problem.obj,
                       x0,
                       method='slsqp',
                       jac=self.problem.grad,
                       tol=1e-5,
                       bounds=bounds,
                       constraints=constraints,
                       options=options,
                       callback=None,
                       )

        #print(out)

        #print(self.calc_constraints(out.x))
        if out.success == True:
            self.feasible = True
        else:
            self.feasible = False
        self.result = out
        self.best_x = out.x
        self.best_f = out.fun
        self.X = out.x
        return self.best_f, self.best_x

    def callback(self, xk, iterations=[]):
        i = len(iterations)
        iterations.append(1)
        if self.verb:
            fval = self.problem.obj(xk)
            self.calc_constraints(xk)
            end = time.time()
            print(f'\r Iteration {i + 1} / {self.maxiter}: Obj: {fval} Feasible: {self.problem.feasible} ' \
                f'max g: {max(self.constr_vals):.4f}')
            print(f"Iteration took: {end - self.start :.2f} s")
            self.start = end

    def feasibility_stop(self, xk):
        
        self.problem.substitute_variables(xk)
        is_feasible = self.problem.feasible        
        return is_feasible
        
        #self.calc_constraints(xk)
        
class COBYLA(SLSQP):

    def solve(self, problem, maxiter=100, x0=[]):
        """
        Solves given problem

        :param problem: problem to be solved
        :param maxiter: number of maximum iterations
        :param x0: initial guess

        :return: fopt, xopt
        """

        self.problem = problem
        eqcons = self._create_eqcons()
        ieqcons = self._create_ieqcons()

        # Create initial guess if one isn't provided
        if not len(x0):
            x0 = []
            for var in problem.vars:
                x0.append(var.ub)

        options = {'maxiter': maxiter,
                   'disp': True,
                   'rhobeg': 1e2}

        constraints = []

        for eqcon in eqcons:
            con = {'type': 'eq', 'fun': eqcon}
            constraints.append(con)

        for ineqcon in ieqcons:
            con = {'type': 'ineq', 'fun': ineqcon}
            constraints.append(con)

        out = minimize(self.problem.obj,
                       x0,
                       method='COBYLA',
                       constraints=constraints,
                       options=options)
        # print(out)
        return out


if __name__ == '__main__':
    from metku.optimization.benchmarks import *

    problem = TenBarTruss('continuous')
    solver = SLSQP()
    xopt, fopt = solver.solve(problem, maxiter=10)

    problem(xopt)



# -*- coding: utf-8 -*-
#  This source code is licensed under the MIT license. See LICENSE in the repository root directory.
#  Copyright 2022. Metku team
#  All rights reserved.
#  Author(s): Jaakko Huusko

from dataclasses import dataclass, field


@dataclass
class SteelMaterial:
    """
    Steel material class

    Default values:
    ---------------
        fy: None, [MPa]
        fu: None, [MPa]
        fy_40: None, [MPa]
        nu: 0.3, []
        E: 210, [GPa]
        rho: 7850 [kg/m³]
    """
    fy: float  # MPa
    fu: float  # MPa
    fy_40: float = None  # Yield strength when thickness over 40 mm
    nu: float = 0.3
    E: float = 210e3  # MPa
    rho: float = 7850  # kg / m³

    def __post_init__(self):
        if self.fy_40 is None:
            self.fy_40 = self.fy

class Steel:
    """
    Default steel materials
    Custom steel material can be created using SteelMaterial -class
    """
    # ---------- S235 -----------
    S235 = SteelMaterial(fy=235, fu=360)

    # ---------- S275 -----------
    S275 = SteelMaterial(fy=275, fu=430)

    # ---------- S355 -----------
    S355 = SteelMaterial(fy=355, fu=510)
    S355J2H = SteelMaterial(fy=355, fu=510)

    # ---------- S420 -----------
    S420 = SteelMaterial(fy=420, fu=500)


class StainlessSteel:
    """
    Default stainless steel materials
    Custom steel material can be created using SteelMaterial -class
    """
    EN14301_coldstrip = SteelMaterial(fy=230, fu=540, E=200)
    EN14301_hotsrip = SteelMaterial(fy=210, fu=520, E=200)
    EN14301_hotplate = SteelMaterial(fy=210, fu=520, E=200)
    EN14301_sections = SteelMaterial(fy=190, fu=500, E=200)

if __name__ == '__main__':
    s235 = Steel.S235
    print(help(Steel))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  This source code is licensed under the MIT license. See LICENSE in the repository root directory.
#  Copyright (c) 2022. Metku team.
#  All rights reserved.

# @Filename : truss2d_testing
# @Date : 2022
# @Project: metku
# @AUTHOR : Jaakko Huusko
from matplotlib import pyplot as plt

import metku.truss2d as t2d
import metku.frame2d as f2d
import numpy as np

def joint_test():
    n_bays = 1
    n_storeys = 1
    L_bay = 10000
    H_storey = 5000

    simple_frame = [n_storeys, n_bays, H_storey, L_bay]

    frame = f2d.Frame2D(simple=simple_frame, create_beams=False, supports='fixed')

    simple_truss = dict(
        H0=H_storey,
        H1=1500,
        H2=2000,
        L1=L_bay / 2,
        H3=1500,
        dx=1000,
        n=16
    )

    truss = t2d.Truss2D(simple=simple_truss)

    for tc in truss.top_chords:
        truss.add(f2d.LineLoad(tc, [-30, -30], 'y', load_id=f2d.LoadIDs.ULS))
    truss.add(f2d.XYHingedSupport([0, simple_truss['H0'] + simple_truss['H1']]))
    truss.add(f2d.YHingedSupport(
        [truss.L1 + truss.L2, simple_truss['H0'] + simple_truss['H3']]))
    truss.generate()
    truss.calculate(load_id=f2d.LoadIDs.ULS)

    truss.H1 = 1000
    truss.H2 = 1500
    truss.H3 = 1300
    truss.dx = 100



    joint = truss.joints[4]
    w1, w2 = joint.webs.values()

    print("Before: ")
    print('e0', joint.e)
    print('g0', joint.g1)
    print('e1', w1.j1.e)
    print('g1', w1.j1.g1)
    print('e2', w2.j1.e)
    print('g2', w2.j1.g1)

    print("theta1", np.degrees(joint.theta1))
    print("theta2", np.degrees(joint.theta2))
    print("-" * 20)
    val = 50
    print("Changed g0 to:", val)
    print("-" * 20)
    joint.g1 = val
    print("After: ")
    print('e0', joint.e)
    print('g0', joint.g1)
    print('e1', w1.j1.e)
    print('g1', w1.j1.g1)
    print('e2', w2.j1.e)
    print('g2', w2.j1.g1)
    print("theta1", np.degrees(joint.theta1))
    print("theta2", np.degrees(joint.theta2))

    truss.plot(show=False)
    for joint in truss.joints.values():
        joint.plot_joint(show=False)
    plt.show()

def displacement_test():
    simple_truss = dict(
        H0=0,
        H1=1200,
        H2=2600,
        H3=1200,
        L1=12000,
        dx=1600,
        n=16
    )

    truss = t2d.Truss2D(simple=simple_truss)

    for tc in truss.top_chords:
        truss.add(f2d.LineLoad(tc, [-50, -50], 'y', load_id=f2d.LoadIDs.ULS))
    truss.add(f2d.XYHingedSupport([0, simple_truss['H0'] + simple_truss['H1']]))
    truss.add(f2d.YHingedSupport(
        [truss.L1 + truss.L2, simple_truss['H0'] + simple_truss['H3']]))
    truss.generate()
    truss.calculate(load_id=f2d.LoadIDs.ULS)
    print(truss.get_deflection(load_id=f2d.LoadIDs.SLS_Charasteristic))
    truss.plot_deflection()

if __name__ == '__main__':
    displacement_test()


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Wed Oct  9 19:29:05 2019

Branch and bound algorithm for optimization

Data structures:
    
    Class BnB .. for running the algorithm
    Class BnBNode .. contains information about branching at given node


@author: kmela
"""

from copy import deepcopy
import numpy as np
from treelib import Tree, Node

from metku.optimization.solvers.optsolver import OptSolver
from metku.optimization.solvers.lp import LP
from metku.optimization.structopt import OptimizationProblem
from metku.optimization.objective import LinearObjective
from metku.optimization.constraints import LinearConstraint, NonLinearConstraint
from metku.optimization.variables import Variable, DiscreteVariable, BinaryVariable, IntegerVariable

CON_TOL = 1e-4

# A class that represents an individual node in a 
# Binary Tree 
class BnBNode: 
    def __init__(self,flb=-np.inf,branched_var=None,lim=None,sense=None,x0=None): 
        """ Constructor 
            input:
                flb .. lower bound for the objective function
                branched_vars .. list of variables that are branched in this node
                lims .. branching limits for the variables
                sense .. branching direction ('<' or '>')
                
            parameters
                xR .. solution of the relaxation
                lb .. f(xR)
        """
        self.x0 = x0
        self.xR = []
        self.lb = flb
        self.var = branched_var
        self.lim = lim
        self.sense = sense
        
    def level(self):
        """ Level of the node: number of branchings """
        return len(self.var)
        
    def add_var_branch(self,var,lim,sense):
        """ Adds variable branching """
        
        self.var.append(var)
        self.lim.append(lim)
        self.sense.append(sense)
        
    def is_feasible(self):
        """ Checks if the lower bounding problem of 
            the node is feasible
        """
        if self.lb > -np.inf:
            return True
        else:
            return False
    
    def info(self):
        """ Prints relevant node data """        
        
        #print(self.sense,self.lims,self.vars)
        print('** BnB Node **')
        print('Branched variable:')
        if self.var is not None:
            print('{0:s} {1:s} {2:4.2f}'.format(self.var.name,self.sense,self.lim))
        #for i, var in enumerate(self.vars):
        #    print('{0:s} {1:s} {2:4.2f}'.format(var.name,self.sense[i],self.lims[i]))
        
    

class BnB(OptSolver):
    """ Class for Branch and Bound tree
        Includes functionalities for running branch and bound algorithm
        
        Data structures:
            Tree (from treelib): branching tree. Each node contains information
            of one branching. When a lower bounding problem is solved at a node,
            the tree is traversed until root to find the branching variables,
            limits and branching directions.
            
            
        
    """
    
    def __init__(self,problem,lb_solver=None):
        """ Constructor
            input:
                problem .. Optimization problem class object that
                            contains the problem data.
                            
                lb_solver .. OptSolver class object that implements the
                             solution method for the continuous relaxation
                            
            attributes:
                nodes .. list of nodes to be explored. Will be dynamically
                         updated during the iterations.
                max_iters .. maximum number of iterations allowed
                xbest .. variable values of the incumbent
                fbest .. objective function value of the incumbent
                
                
            self.constr_vals = np.array([])
            self.X = np.array([])
            self.problem = None
            self.fvals = []
            self.xvals = []
            self.best_f = np.inf
            self.best_x = None
        """
        super().__init__()
        
        self.created_nodes = 0
        self.problem = problem
        self.lb_solver = lb_solver
        # Initialize empty list of nodes
        # These are node identifier values correspondng to tree nodes
        self.nodes = []
        
        self.max_iters = 20
        
        self.orig_lb = []
        self.orig_ub = []
        
        self.set_original_bounds()
        
        # treelib Tree object that contains information about the 
        # branchings
        self.tree = Tree()
        #self.best_x = []
        #self.fbest = np.inf
        
    def add_node(self,node,tag=None,identifier=None,parent=None):
        """ Adding a node means creating a new node to the B&B tree
            as well as appending the new node to the list of unexplored nodes
        """
        #self.nodes.append(node)
        self.tree.create_node(tag,identifier,parent,data=node)
        
        self.nodes.append(identifier)
        
    def remove_node(self,node_id):
        """ Removes the node with identifier 'node_id' from
            B&B tree as well as from the list of unexplored nodes
        """
        
        self.tree.remove_node(node_id)
        
        try:
            self.nodes.remove(node_id)
        except ValueError:
            print("Node removal did not succeed.")
            pass
        
    def set_original_bounds(self):
        """ Stores the original variable bounds """
        for var in self.problem.vars:
            self.orig_lb.append(deepcopy(var.lb))
            self.orig_ub.append(deepcopy(var.ub))
            
    def set_variable_bounds(self):
        """ Stores the original variable bounds """
        for i, var in enumerate(self.problem.vars):            
            var.lb = deepcopy(self.orig_lb[i])
            var.ub = deepcopy(self.orig_ub[i])            
    
    def max_level_nodes(self):
        """ Finds the maximum level in the list of nodes, and returns
            the corresponding nodes
        """
        level_max = -1
        max_nodes = []    
        
        for node in self.nodes:            
            if node.level() > level_max:
                """ If the level of current node is larger than the current
                    maximum, set the level of the node as the maximum and
                    set the list of nodes to the current node
                """
                level_max = node.level()
                max_nodes = [node]
            elif node.level() == level_max:
                """ If the level of the current node equals the current maximum,
                    add the node to the list of nodes
                """
                max_nodes.append(node)
            
        return level_max, max_nodes
    
    def min_level_nodes(self):
        """ Finds the minimum level in the list of nodes, and returns
            the corresponding nodes
        """
        level_min = np.inf
        min_nodes = []
        
        for node in self.nodes:          
            if node.level() < level_min:
                """ If the level of current node is smaller than the current
                    minimum, set the level of the node as the minimum and
                    set the list of nodes to the current node
                """
                level_min = node.level()
                min_nodes = [node]
            elif node.level() == level_min:
                """ If the level of the current node equals the current minimum,
                    add the node to the list of nodes
                """
                min_nodes.append(node)
            
        return level_min, min_nodes
    
    def select_node(self,strategy="depth_first"):
        """ Selects node 
            input:
                strategy .. 
                'depth_first': choose the node from among those with lowest level
                'breath_first: chooses the node from among those with highest level
            
            output:
                treelib.node object
        """
                 
    
        if strategy == 'depth_first':
            """ Find the deepest of the unexplored nodes """
            nodes = []
            max_depth = 0             
            for node in self.nodes:
                node_depth = self.tree.depth(node)
                if node_depth > max_depth:
                    nodes = [self.tree.get_node(node)]
                    max_depth = node_depth
                elif node_depth == max_depth:
                    nodes.append(self.tree.get_node(node))
                    
            #depth = self.tree.depth()        
            #nodes = list(self.tree.filter_nodes(lambda x: self.tree.depth(x)==depth))
            #print(nodes)
            #level, nodes = self.max_level_nodes()
            
        elif strategy == 'breath_first':
            """ Find the nodes with smallest level """
            level, nodes = self.min_level_nodes()
        
        
        print('Node selection:')
        print('Strategy: {0}'.format(strategy))        
        
        """ From among the set of nodes chosen according to the strategy,
            choose the one with the smallest lower bound of the objective
            function
        """
        fmin = np.inf
        for node in nodes:
            if node.data.lb < fmin:
                fmin = node.data.lb
                selected_node = node
        
        try:        
            self.nodes.remove(selected_node.identifier)
        except ValueError:
            print("**** Selected node not found in the list of nodes!*****")
            print(selected_node.identifier)
            print(self.nodes)
                
        
        """ Apply variable bounds included in the node """
        
        """ Set variable bounds to their original values """
        self.set_variable_bounds()
        
        #for var in self.problem.vars:
        #    print("{0} <= {1:s} <= {2}".format(var.lb,var.name,var.ub))
        
        #subtree = self.tree.expand_tree(node.identifier,reverse=True)
        
        
        
        #print(subtree)
        
        #for node in subtree:
        for sub_node in list(self.tree.rsearch(selected_node.identifier)):
            if not self.tree[sub_node].is_root():
                bnode = self.tree[sub_node]            
                if bnode.data.sense == '<':
                    bnode.data.var.ub = min(bnode.data.var.ub,bnode.data.lim)
                else:
                    bnode.data.var.lb = max(bnode.data.var.lb,bnode.data.lim)                
        
        #for var in self.problem.vars:
        #    if var.lb == var.ub:
        #        var.lock(var.lb)            
                 
        #    print("{0} <= {1:s} <= {2}".format(var.lb,var.name,var.ub))
        
        
        selected_node.data.info()
        
        return selected_node
    

    def lower_bound(self, node, x0=None):
        """ Finds a lower bound for the problem at the 'node' by
            solving a relaxation of the problem and imposing the
            branching limits given by the node.
        """
        
        """ Set the variable bounds to the problem corresponding
            to the branches in the node
        """
                
        """ Set variable bounds to their original values """
        
        """
        self.set_variable_bounds()
        
        for var in self.problem.vars:
            print("{0} <= {1:s} <= {2}".format(var.lb,var.name,var.ub))
        
        #subtree = self.tree.expand_tree(node.identifier,reverse=True)
        
        
        
        #print(subtree)
        
        #for node in subtree:
        for sub_node in list(self.tree.rsearch(node.identifier)):
            if not self.tree[sub_node].is_root():
                bnode = self.tree[sub_node]            
                if bnode.data.sense == '<':
                    bnode.data.var.ub = min(bnode.data.var.ub,bnode.data.lim)
                else:
                    bnode.data.var.lb = max(bnode.data.var.lb,bnode.data.lim)                
        
        for var in self.problem.vars:
            print("{0} <= {1:s} <= {2}".format(var.lb,var.name,var.ub))
        """    
              
                
        """ Run 'solver'. 
            Solver should be LP or NLP solver
            Examples:
                a) few iterations of SLP
                b) NLP method to optimality -> heuristic, if no convexification is done
    
            If problem is infeasible, the node can be fathomed.
            If the problem is feasible, update the lower bound and set xR
        """
        nvars = self.problem.nvars()
        locked_vars = []
        locked_nd = []
        for i,var in enumerate(self.problem.vars):
            if abs(var.ub-var.lb) < 1e-6:
                var.lock(var.ub)
                locked_vars.append(var)
                locked_nd.append(i)

        if len(locked_nd) > 1:
            locked_nd.reverse()
        
        for n in locked_nd:
            x0 = np.delete(x0,n)        
    
        print('x0 = ',x0)
    
        if x0 is not None:
            flb, xlb = self.lb_solver.solve(self.problem,x0=x0,verb=True)
        else:
            flb, xlb = self.lb_solver.solve(self.problem)

        print('xlb = ',xlb)

        xnew = np.zeros(nvars)
        
        i = 0
        for j, var in enumerate(self.problem.all_vars):
            if var.locked:
                xnew[j] = var.value
            else:
                xnew[j] = xlb[i]
                i +=1

        #print(self.lb_solver.result)
        if self.lb_solver.result.success == True:
            self.tree[node.identifier].data.lb = flb
            self.tree[node.identifier].data.xR = xnew   
        elif self.lb_solver.feasible:
            self.tree[node.identifier].data.lb = flb
            self.tree[node.identifier].data.xR = xnew
            
        """ unlock variables at their bounds in this node """
        for var in locked_vars:
            var.unlock()
        
        return self.lb_solver.feasible
        
    
    def upper_bound(self,node):
        """ Try to find feasible solution using a heuristic etc. """
        N = self.problem.discrete_neighborhood(node.data.xR, k=2)
        f0 = deepcopy(self.best_f)
        f_new = None        
        for xi in N:
            
            gxi = self.problem.eval_cons(xi)            
            
            if all(gxi<=self.problem.con_tol):
                #print("Upper bounding found feasible solution")
                #print(self.problem.eval_cons(xi))
                """ Feasible point found """
                f_new = self.problem.obj(xi)
                
                if f_new < self.best_f:
                    print("Better feasible solution found:")
                    print("Current: {0:4.2f}, New: {1:4.2f}".format(self.best_f,f_new))
                    print("Current: ",self.best_x)                    
                    print("New: ",xi)
                    #print("Constraints: ",self.problem.eval_cons(xi))
                    
                    self.best_f = f_new
                    self.best_x = list(xi)
        
        #if f_new is not None and fnew < f0:
        #    """ Feasible solution was improved so all nodes with
        #        Larger lower bound can be removed from the search tree
        #    """
            #self.remove_node(node.identifier)
        #    pass
        
                                        
        
    
    
    def variables_at_discrete_values(self,node):
        """ Checks, whether all discrete variables have discrete
            values at the solution of the relaxation at given node.
        """ 
        xR = node.xR
            
        #print(xR)
        res = []
        
        for i, var in enumerate(self.problem.vars):
            res.append(var.is_allowed_value(xR[i]))
            
        
        return res
        
            
    def branch(self,node,strategy='max_int_violation'):
        """ Branching of a node """
                
        xR = node.data.xR
        #print(xR)
        
        if strategy == 'max_int_violation':
            max_violation = 0
            branching_var = None
            for i, var in enumerate(self.problem.vars):                
                violation_i = var.discrete_violation(xR[i])
                print(var.name,var.branch_priority,violation_i)
                if branching_var is None:
                    if violation_i > max_violation:
                        max_violation = violation_i
                        branching_var = var                    
                        xr = xR[i]
                elif var.branch_priority > branching_var.branch_priority:
                    if violation_i > 0:
                        max_violation = violation_i
                        branching_var = var                    
                        xr = xR[i]
                elif var.branch_priority == branching_var.branch_priority:
                    if violation_i > max_violation:
                        max_violation = violation_i
                        branching_var = var                    
                        xr = xR[i] 
        
        var_ndx = self.problem.vars.index(branching_var)
        
        """ Create branching values """
        low_branch_value = branching_var.smaller_discrete_value(xr)
        high_branch_value = branching_var.larger_discrete_value(xr)
        
        """ Create new nodes """
        x01 = deepcopy(xR)
        x01[var_ndx] = low_branch_value
        new_node1 = BnBNode(flb=deepcopy(node.data.lb),branched_var=branching_var,lim=low_branch_value,sense='<',x0=x01)
        tag1 = branching_var.name + ' <= ' + str(low_branch_value)
        id1 = self.created_nodes + 1
        
        x02 = deepcopy(xR)
        x02[var_ndx] = high_branch_value
        new_node2 = BnBNode(flb=deepcopy(node.data.lb),branched_var=branching_var,lim=high_branch_value,sense='>',x0=x02)
        tag2 = branching_var.name + ' >= ' + str(high_branch_value)
        #id2 = str(self.tree.size()+1)
        id2 = self.created_nodes + 2
        self.created_nodes += 2

        self.add_node(new_node1,tag1,id1,parent=node)
        self.add_node(new_node2,tag2,id2,parent=node)
        
        #self.tree.show()
        
    def feasibility_tightening(self):
        """ Feasibility-based bounds tightening
            Based on linear constraints only        
        """
        print("Feasibility-based bounds tightening")
        xlb, xub = self.problem.var_bounds()
        
        """ Do bounds tightening for linear constraints """
        for con in self.problem.cons:
            if isinstance(con,LinearConstraint):
                ax_min = np.minimum(con.a*xub,con.a*xlb)
                #axL = con.a*xlb
                for i, var in enumerate(self.problem.vars):
                    if abs(con.a[i]) > 1e-9:
                        bnd = 1/con.a[i]*(con.b-(sum(ax_min)-ax_min[i]))
                    if con.a[i] > 0:                        
                        """ x[i] <= 1/a[i]*(b[j] - sum min(a[i,j]xub[j],a[i,j]xlb[j])) """
                        var.ub = min(var.ub,bnd)
                        
                        if isinstance(var,IntegerVariable):
                            var.ub = np.floor(var.ub)
                        elif isinstance(var,DiscreteVariable):
                            var.ub = max(np.array(var.values)[np.array(var.values)<=var.ub])
                    elif con.a[i] < 0:
                        var.lb = max(var.lb,bnd)
                
                        if isinstance(var,IntegerVariable):
                            var.lb = np.ceil(var.lb)
                        elif isinstance(var,DiscreteVariable):
                            var.lb = min(np.array(var.values)[np.array(var.values)>=var.lb])
        
        xlb, xub = self.problem.var_bounds()
        #print(xlb,xub)
    
    def optimality_tightening(self):
        """ Optimality-based bounds tightening
            Solve a series of LP-problems to find
            minimum and maximum values for variables
        """
        print("Optimality-based bounds tightening:")
        
        for i, var in enumerate(self.problem.vars):
            lobj = np.zeros(self.problem.nvars())
            """ Formulate linear problem """
            
            LPprob = OptimizationProblem(name="LP_tight",constraints = [],variables=self.problem.vars)            
            lobj[i] = 1
            lObjective = LinearObjective("LP_tight_obj",c=lobj,obj_type="MIN",problem=LPprob)
            
            
            LPprob.add(lObjective)
            
            
            for con in self.problem.cons:            
                if isinstance(con,LinearConstraint):
                    LPprob.add(con)
            
            
            solver = LP("gurobi")
            solver.solve(LPprob)
            
    
            #LPprob(solver.X)
            if solver.feasible == True: 
                #print("Upper bounding problem feasible.")                   
                if solver.X[i] > self.problem.vars[i].lb:
                    print("Increasing lower bound of {0:s}".format(self.problem.vars[i].name))
                    print("From {0:4.2f} to {1:4.2f}".format(self.problem.vars[i].ub,solver.X[i]))
                    self.problem.vars[i].lb = solver.X[i]
            
            
            #print(solver.X)
            
            LPprob.obj.obj_type = "MAX"
            solver.solve(LPprob)
            
            
            if solver.feasible == True:        
                #print("Lower bounding problem feasible.")
                if solver.X[i] < self.problem.vars[i].ub:
                    print("Lowering upper bound of {0:s}".format(self.problem.vars[i].name))
                    print("From {0:4.2f} to {1:4.2f}".format(self.problem.vars[i].ub,solver.X[i]))
                    self.problem.vars[i].ub = solver.X[i]
                    
            #print(solver.X)
            
    
    def pre_process(self,node):
        """ Pre-processing of a node """
        self.feasibility_tightening()
        self.optimality_tightening()
        
        feasible = True
        
        for var in self.problem.vars:
            if var.lb > var.ub:
                feasible = False
                break
        
        return feasible
    
    def post_process(self,node):
        """ Post-processing of a node """
        pass    
    
    def solve(self, problem, x0=None, maxiter=200, verb = 0):
        """ Runs the Branch and Bound algorithm """
        
        self.problem = problem
        
        self.max_iters = maxiter
        
        iteration = 0
        
        # Initialize tree
        if len(self.tree.nodes) == 0:
            root = BnBNode(x0=x0)
            self.add_node(root,"Root","root")
        
        while iteration <= self.max_iters and len(self.nodes)>0:
        #while iteration <= 0 and len(self.nodes) > 0:
            self.tree.show()    
            if verb > 0:
                print("*********************************")
                print("****** Begin iteration {0} ******".format(iteration))
                print("Unexplored nodes: {0}".format(len(self.nodes)))
                print(self.nodes)
                
            """ Choose node to be solved
                This is treelib.Node object
            """
            node = self.select_node(strategy="depth_first")            
            
            if verb > 0:
                print("Incumbent: {0:4.2f}".format(self.best_f))
                print("Best design: ", self.best_x)
            
            # Preprocessing:
            # Tighten bounds of the problem
            feasible_pre_processing = self.pre_process(node)
            
            if feasible_pre_processing:     
                if verb > 0:
                    print("Feasible pre-processing.")
                # Find lower bound for the node                
                feasible_lower_bound = self.lower_bound(node,x0)

                
                if feasible_lower_bound:
                    if verb > 0:
                        print("Lower bounding problem is feasible:")
                        print("Lower bound obtained: {0:4.2f} (f_incumbent = {1:4.2f})".format(node.data.lb,self.best_f))                
                
                    if node.data.lb > self.best_f:
                        # Node can be fathomed
                        print("Node fathomed because lower bound is greater than incumbent.")
                        self.remove_node(node.identifier)
                    else:
                        # Try to find a feasible solution in the node, to be
                        # used as an upper bound for the problem.
                        self.upper_bound(node)
                    
                        # Tighten the bounds based on the solution found
                        self.post_process(node)
                    
                        disc = self.variables_at_discrete_values(node.data)                
                        # Here, the relaxation is feasible
                        
                            
                        if all(disc):
                            """ If all discrete variables have discrete values                    
                                in the solution xR, the node can be fathomed
                                If the node provides an improvement to the
                                best known solution, make it as the incumbent.
                            """
                            print("All discrete variables at allowed values!")
                            if node.data.lb < self.best_f:
                                self.best_f = node.data.lb
                                self.best_x = node.data.xR
                            
                            self.remove_node(node.identifier)
                        else:
                            """ In this case, some of the discrete variables
                                have non-integer/non-discrete values at xR.
                                Then, the node is branched.
                            """
                            print("Branching")
                            self.branch(node)                    
                
                else:
                    print("Lower bounding problem is infeasible.")
                    """ Remove node from branching tree """   
                    self.remove_node(node.identifier)
                    #self.tree.remove_node(node.identifier)
            else:
                print("Bounds tightening resulted detected infeasiblity.")
                """ Remove node from branching tree """   
                self.remove_node(node.identifier)
                    
            iteration += 1
        
        # Wrap up: set original variable bounds
        self.set_variable_bounds()
    
        return self.best_f, self.best_x


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
class PointLoad(Load):
    """ Class for point loads (forces and moments)

        Parameters:
        ------------
        :param sid:  load id
        :param node: node subjected to the load
        :param v: load vector [Fx, Fy, Mz] [kN, kN, kNm]
        :param f: scaling factor

        :type sid: int
        :type node: FEMNode
        :type v: list
        :type f: float

        Variables:
        --------------
        :ivar sid:  load id
        :ivar node: node subjected to the load
        :ivar v: load vector [Fx, Fy, Mz] [kN, kN, kNm]
        :ivar f: scaling factor

        :vartype sid: int
        :vartype node: FEMNode
        :vartype v: list
        :vartype f: float

    """

    def __init__(self, sid, node, v, f):
        Load.__init__(self, sid)

        self.node = node
        """ node subjected to the load"""
        self.v = v
        """ load vector"""
        self.f = f
        """ scaling factor"""

    def load_and_dofs(self):
        """ Returns the loads to be inserted to the global load vector
            and the corresponding global degrees of freedom

            Returns:
            ---------

            :return: load to be inserted to the global load vector and
                    the corresponding global degrees of freedom
            :rtype: (int, int)

        """

        """ Load vector """
        F = self.f * np.array(self.v)

        """ Degrees of freedom """
        dofs = np.array(self.node.dofs)

        """ Find free degrees of freedom (those with numbering
                                          greater than -1)
        """
        nzero_dofs = dofs >= 0

        """ Get free degrees of freedom and the corresponding part
            of the load vector
        """
        d = dofs[nzero_dofs]
        Fnz = F[nzero_dofs]

        return Fnz, d



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Wed Jul  1 17:03:17 2020

Truss optimization

@author: kmela
"""

import numpy as np
import re


from metku.frame2d.frame2d import SteelBeam
from metku.raami.raami_simple_truss import SimpleTruss, three_bar_truss, ten_bar_truss
from metku.sections.steel import SteelSection

from metku.optimization.structopt import OptimizationProblem
from metku.optimization.variables import Variable, DiscreteVariable, IndexVariable, BinaryVariable
from metku.optimization.constraints import NonLinearConstraint, LinearConstraint
from metku.optimization.objective import ObjectiveFunction, LinearObjective

from metku.sections.steel.catalogue import ipe_profiles, shs_profiles, hea_profiles
from metku.sections.steel import ISection, WISection, RHS, SHS
from metku.optimization.solvers.lp import LP
from metku.optimization.solvers.milp import MILP



class TrussOptimization(OptimizationProblem):
    """
    Class for general structural problem
    
    13.7.2020
    
    Tehtävätyypit:
        'sizing'
        'geometry'
        'topology'
        
    Tehtävät, joita on tarkoitus ratkaista:
        'sizing':
            1) jatkuvina muuttujina vain poikkipinta-alat
            2) diskreetteinä muuttujina vain poikkipinta-alat
            3) profiilit katalogista: MILP -formulointi
            4) profiilit katalogista: NAND -formulointi
            5) SAND-formulointi
    
    Rakenne on lähtökohtaisesti nivelpäisten sauvojen joukko, jossa sauvat
    kantavan vain normaalivoimaa ja joissa siirtyminä ovat vain solmusiirtymät.
    
    Tätä voitaisiin vielä laajentaa ristikoille, jotka analysoidaan palkkielementtien
    avulla.
    
    Pitänee tehdä erilliset tiedostot erilaisille rakenteille:
        'PortalTruss'
        'ArchedTruss'
        'BeamColumnFrame'
    
    jne.

    """

    def __init__(self,
                 name="Truss Optimization Problem",
                 problem_type = "sizing",
                 variables=[], constraints=[], objective=None, \
                 gradient=None, hess=None, structure=None, profiles=None):
        
        """         problem_type="sizing",
                 formulation='NAND',
                 structure=None,
                 var_groups=None,
                 con_groups=None,
                 constraints=None,
                 profiles=None,
                 obj=None):
        """
        """ Constructor 
            Parameters:
            ------------
            :param name: Name of the problem (string)
            :param structure: Structure to be optimized (Frame2D)
            :param var_groups: grouping of design variables
            :param con_groups: constraint groups
            :param constraints: constraints of the problem
            :param profiles: available profiles
            :param obj: objective function
    
            :type name: string
            :type structure: Frame2D member?
            :type var_groups: list or dict?
            :type con_groups: list or dict?
            :type constraints: list or dict?
            :type profiles: list or dict?
            :type obj: callable or string?
    
    
            Variables:
            ----------
            :ivar vars: list of variables
            :ivar var_groups:
            :ivar con_groups:
            :ivar profiles:
            :ivar constraints:
            :ivar
        
        """
        
        """ 'constraints' can be a dict or list of constraint functions """
        if isinstance(constraints,dict):
            constraint_funs = []
        else:
            constraint_funs = constraints
        super().__init__(name,variables, constraint_funs, objective, gradient, hess, structure, profiles)
        
        self._ptype = problem_type
        #self.vars = []
        var_groups = []
        con_groups = []
        self.var_groups = var_groups
        self.con_groups = con_groups
        
        """ If no profile list is given, use the default SHS profiles """
        if profiles is None:
            profiles = list(shs_profiles.keys())
            
        self.profiles = profiles
        
        """ Generate variables.
            By default, index variables are generated. This can be overridden
            by var_groups, which gives the variable data.
        """
        #self.create_variables()
        
        """ Generate constraints 
            By default, cross-section resistance constraints are generated.
            This can be overridden by 'constraints' dict.
        """
        if constraints is None or len(constraints) == 0:
            constraints = {
                'stress': True,
            }
        
        self.constraints = constraints
    
        self.stress_cons = []
        self.displacement_cons = []
    
        self.create_constraints(**self.constraints)
        #self.group_constraints()
        if objective is not None:
            self.obj = objective
        else:
            self.create_objective()

    @property
    def available_constraints(self):
        """
        Returns all available constraints and their corresponding keyword
        :return:
        """
        return dict(stress=False,
                    strength=False,
                    buckling=False)
    
    @property
    def available_formulations(self):
        """
        Returns all available formulations and their corresponding keyword
        :return:
        """
        return dict(lp=False,
                    nand=False,
                    sand=False,
                    discrete=False,
                    sizing=False,
                    geometry=False,
                    topology=False)
    
    def create_variables(self,var_type="area",var_cat="continuous",**kwargs):
        """
        Creates design variables
            :param var_type: variable type. Possible values: 'area', 'profile', 'index'
            :param var_cat: variable category. Posstible values: 'continuous', 'discrete', 'integer'
            :param **kwargs: additional parameters to complete the definition, depending on the variable type and category
        """
        self.vars.clear()

        if var_type == "area":
            var_prop = 'A'
            lb = 1e-3
            ub = 1e5
            if var_cat == "continuous":
                for key, value in kwargs.items():
                    if key == 'lb':
                        lb = value
                    elif key == 'ub':
                        ub = value
                
                x0 = 0.5*(lb+ub)

        # If no grouping is given, create a group for each member
        if self.var_groups is None or len(self.var_groups) == 0:
            self.var_groups = []
            for mem in self.structure.members.values():
                group = {
                    "name": f"{mem.mtype.capitalize()} {mem.mem_id}",
                    "objects": [mem],
                    "value": x0,
                    "profiles": self.profiles,
                    "property": var_prop,
                    "var_type": var_cat,
                    "lb": lb,
                    "ub": ub
                }
                self.var_groups.append(group)

        #print(self.var_groups)
        # Main loop
        for i, group in enumerate(self.var_groups):
            print(group["var_type"])
            if group["var_type"] == 'discrete':
                var = DiscreteVariable(
                    name=group['name'],
                    values=group['values'],
                    value=group['value'],
                    target={"property": group["property"],
                            "objects": group["objects"]}
                )
                self.add(var)

            elif group["var_type"] == 'continuous':
                # Multiple properties
                if "properties" in group.keys():
                    bounds = group["bounds"]
                    if len(bounds) != len(group["properties"]):
                        raise ValueError(
                            "There must be same number of bounds as properties!"
                            f"{group['bounds']} != {group['properties']}")
                    for bounds, prop, value in zip(group["bounds"],
                                                   group["properties"],
                                                   group["values"]):
                        lb, ub = bounds
                        var = Variable(
                            name=group['name'],
                            lb=lb,
                            ub=ub,
                            value=value,
                            target={"property": prop,
                                    "objects": group["objects"]}
                        )
                        self.add(var)

                # Single property
                else:
                    var = Variable(
                        name=group['name'],
                        lb=group['lb'],
                        ub=group['ub'],
                        value=group['value'],
                        target={"property": group["property"],
                                "objects": group["objects"]}
                    )
                    self.add(var)

            elif group["var_type"] == 'index':
                var = IndexVariable(
                    name=group['name'],
                    value=group['value'],
                    values=group['values'],
                    target={"property": group["property"],
                            "objects": group["objects"]}
                )
                self.add(var)

            elif group["var_type"] == 'binary':
                # TODO
                pass
            else:
                raise ValueError("var_type must be either 'discrete',"
                                 " 'continuous', 'index' or 'binary")
    
    def create_objective(self,scaling=10.0):
        """ By default, weight of the structure is taken as the
            objective function.
        """
        def obj_fun(x):
            """ Variable substitution is not shown """
            self.substitute_variables(x)
            return self.structure.weight/scaling

        obj = ObjectiveFunction(name="Weight",
                                obj_fun=obj_fun,
                                obj_type='MIN')
        self.add(obj)
        
    def create_constraints(self,
                           stress=False,
                           strength=False,
                           buckling=False,
                           members=None
                           ):
        """
        Creates constraints

        NOTE! The name of these kwargs must match the key in the cons -dict
        created in constraint function creator method
        e.g.
        CORRECT:
        cons = {'compression': comp_fun, 'tension': ten_fun}
        FALSE:
        cons = {'comp_func': comp_fun, 'tension_con': ten_fun}

        :param stress:
        :param strength:
        :param buckling:
        :param members:
        :return:
        """

        if members is None:
            members = self.structure.members.values()
            # Clear constraint list
            self.cons.clear()

        for mem in members:
            if stress:
                stress_cons = self.stress_constraints(mem)
                for key, val in stress_cons.items():                    
                    con = NonLinearConstraint(
                        name=f"{key}: {mem.mem_id}",
                        con_fun=val,
                        con_type='<',
                        fea_required=True,
                        vars=mem)
                    self.add(con)
                    self.stress_cons.append(con)
            
            if strength:
                new_cons = self.strength_constraints(mem)
                for key, val in new_cons.items():                    
                    con = NonLinearConstraint(
                        name=f"{key}: {mem.mem_id}",
                        con_fun=val,
                        con_type='<',
                        fea_required=True,
                        vars=mem)
                    self.add(con)
            
            if buckling:
                new_cons = self.buckling_constraints(mem,buckling)
                for key, val in new_cons.items():
                    con = NonLinearConstraint(
                    name=f"{key}: {mem.mem_id}",
                    con_fun=val,
                    con_type='<',
                    fea_required=True,
                    vars=mem)
                
                self.add(con)
            

    
    def stress_constraints(self, mem, max_stress=None):
        """
        Axial stress constraints

        :param mem: FrameMember object
        :param max_stress: maximum allowable stress

        :return: dict of created functions
        """
        
        if max_stress == None:
            max_stress = mem.material.fy

        def compression(x):
            return -mem.ned / (max_stress*mem.A) - 1

        def tension(x):            
            return mem.ned / (max_stress*mem.A) - 1

        cons = {
            'compression': compression,
            'tension': tension,
        }

        return cons
    
    def strength_constraints(self, mem):
        """
        Constraints for member strength against axial force

        :param mem: FrameMember object

        :return: dict of created functions
        """
                
        def compression(x):
            self.substitute_variables(x)
            return -mem.ned / mem.NRd - 1

        def tension(x):            
            self.substitute_variables(x)
            return mem.ned / mem.NRd - 1

        cons = {
            'compression': compression,
            'tension': tension,
        }

        return cons

    def buckling_constraints(self, mem, buckling_type="EN 1993"):
        """
        Member buckling constraints
        
        :param mem: FrameMeber object
        :param buckling_type: 'ec3', 'euler', or something else
        
        :return: dict of constraints
        """
        
        def buckling(x):
            self.substitute_variables(x)
            if buckling_type == 'EN 1993':
                NRd = mem.member.NbRd[0]
            elif buckling_type == 'euler':
                NRd = mem.member.ncrit()[0]
            
            return -mem.ned/NRd - 1
        
        cons = {f"buckling ({buckling_type})": buckling}
        
        return cons
    
    def displacement_constraint(self,node,dof='x',ulb=None,uub=None,load_case=None):
        """

        Parameters
        ----------
        node : FEMNode 
            Nodes whose displacement is to be constrained.
        dofs : LIST, optional
            Degrees of freedom to be constrained. The default is ['x','y'].
        ulb : LIST, optional
            Lower bound values for the displacements. The default is [-100,-100].
        uub : LIST, optional
            Lower bound values for the displacements. The default is [100,100].

        Returns
        -------
        None.
        """
        
        # Displacement with respect to x coordinate is the first
        # element in the list of nodal displacements.
        if dof == 'x':
            dof_nd = 0
        elif dof == 'y':
            dof_nd = 1
        
        if load_case is None:
            load_case = self.structure.load_ids[0]
        
        def disp_upper_bound(x):
            # Constraint is of the form u - uub <= 0
            self.substitute_variables(x)
            
            u = node.u[load_case][dof_nd]            
            
            if uub > 0.0:
                res = u/uub-1
            elif uub < 0.0:
                res = 1-u/uub
            else:
                res = u
                
            print(u)
            return res
        
        def disp_lower_bound(x):
            # Constraint is of the form u >= ulb -> -u+ulb <= 0
            self.substitute_variables(x)
            
            u = node.u[load_case][dof_nd]
            print(u)
            if ulb < 0.0:
                res = u/ulb-1
            elif ulb > 0.0:
                res = 1-u/ulb
            else:
                res = -u
                                    
            return res
                        
        if ulb is not None:
            #cons['f"Displacement lower bound"'] = disp_lower_bound
            con = NonLinearConstraint(
                        name=f"u_{dof}[{node.nid}][load_case] >= {ulb}",
                        con_fun=disp_lower_bound,
                        con_type='<',
                        fea_required=True)
            self.add(con)
            self.displacement_cons.append(con)
        
        if uub is not None:
            #cons['f"Displacement upper bound"'] = disp_upper_bound
            con = NonLinearConstraint(
                        name=f"u_{dof}[{node.nid}][load_case] <= {uub}",
                        con_fun=disp_upper_bound,
                        con_type='<',
                        fea_required=True)
            self.add(con)
            self.displacement_cons.append(con)
        
        #return cons
            
    
    def fsd(self,verb=False):
        """
        Find fully stressed design for the truss.
        It is assumed that the variables are cross-sectional areas and
        there are only stress constraints

        Returns
        -------
        None.

        """
        kmax = 10
        k = 0
        
        if verb:
            print("*** FSD iteration ***")            
                
        while k < kmax:
            print("  Iteration {0:3g}".format(k))
            k += 1
            for load_case in self.structure.load_ids:
                # Calculate response of the structure
                # for this load case
                self.fea(load_case)
                
                if verb:
                    print("   Load case: {0:s}".format(load_case))
                            
                for var in self.vars:
                    # For each variable, go through
                    # all members that the variable is targeted at.
                    for mem in var.target['objects']:
                        # Run update rule:
                        # 1. get member force
                        # 2. update variable value
                        N = mem.member.NEd
                        new_val = N/mem.member.fy()
                        
                        new_val = min(max(new_val,var.lb),var.ub)
                        
                        if verb:
                            print("{0:s}(old) = {1:5.3f}, N = {2:5.3f}, NRd = {4:5.3f}, {0:s}(new) = {3:5.3f}".format(var.name,var.value,N,new_val,var.value*mem.member.fy()))
                                            
                        
                        var.substitute(new_val)
        
        #self.print_vars()
    
    def scale_design(self,x):
        """
        Scales the design 'x' such that stress and discplacement constraints
        are satisfied. It is assumed that the design variables are the cross sectional
        areas.

        Parameters
        ----------
        x : LIST
            Design to be scaled.

        Returns
        -------
        None.

        """
        mu_U = np.inf
        mu_L = -np.inf
                
        for i, var in enumerate(self.vars):
            # Upper bound for the scaling factor is obtained from
            # design variable bounds
            mu_U = min(mu_U,var.ub/x[i])
        
            # Design variable bounds affect the lower bound too.
            mu_L = max(mu_L,var.lb/x[i])
        
        for con in self.displacement_cons:
            g = con(x)
            
            mu_L = max(mu_L,g+1)
        
        for con in self.stress_cons:
            g = con(x)
            
            mu_L = max(mu_L,g+1)
            
        print(mu_L,mu_U)
        
        if mu_L > mu_U:
            print("Design cannot be scaled.")
            mu_L = 1.0
        
        return mu_L*x
            
        
def truss_lp(truss,Alb=0,Aub=1e9,slb=None,sub=None):
    """ Creates minimum weight LP problem for the truss 
        Design variables:
            1) Cross-sectional areas of the members
            2) Axial forces of the members, for each load combination
        
        Constraints:
            1) Equilibrium equations, for each load combination
            2) Force constraints, tension and compression
            3) Box constraints for cross-sectional areas
    """
        
    """ Create variables """
    dvars = []
    c = []
    
    nmem = len(truss.members)
    
    """ Initialize cross-sectional area and stress bounds """
    if not (isinstance(Alb,np.ndarray) or isinstance(Alb,list)):        
        Alb = Alb*np.ones(nmem)
    
    if not (isinstance(Aub,np.ndarray) or isinstance(Aub,list)):
        Aub = Aub*np.ones(nmem)
        
    if slb is None:
        slb = [mem.fy for i, mem in t.members.items()]
    
    if sub is None:
        sub = [mem.fy for i, mem in t.members.items()]
    
    if not (isinstance(slb,np.ndarray) or isinstance(slb,list)):        
        slb = slb*np.ones(nmem)
    
    if not (isinstance(sub,np.ndarray) or isinstance(sub,list)):
        sub = sub*np.ones(nmem)
    
    
    
    """ Design variables: member areas 
        The assumption here is that each member has its own variable.
        If member grouping is used, this needs to be changed according to the
        grouping.
        
        Also the stress constraints must be changed to take into account groups.
    """
    for i, mem in truss.members.items():
        dvars.append(Variable(name="A{0:.0g}".format(i),lb=Alb[i],ub=Aub[i],
                                  target={"objects":[mem],"property":["A"]},
                                  value=mem.A))
        c.append(mem.length)
    
    """ Design variables: member forces """
    for lc in truss.load_cases:
        for i, mem in truss.members.items():
            dvars.append(Variable(name="N{0:.0g}{1:.0g}".format(lc,i),lb=-1e9,ub=1e9,
                                  target=None,
                                  value=1))
            c.append(0)
            
    
    """ Initialize optimization problem with the created variables """
    lp = OptimizationProblem(name="Truss LP",variables=dvars,structure=truss)
    
    """ Objective function: truss weight """
    obj = LinearObjective("Weight", c)
    lp.add(obj)
    
    """ Constraints: equilibrium equations """
    B = truss.f.statics_matrix()
    ndof = truss.f.nfree_dofs
        
    for j, lc in enumerate(truss.load_cases):
        """ Load vector for the load case, including only unsupported dofs """
        p = np.delete(truss.f.global_load_vector(lc),truss.f.supp_dofs)
        """ a is a placeholder for the coefficient vector of a linear constraint """
        a = np.zeros(lp.nvars())
        for i, (b,pi) in enumerate(zip(B,p)):
            """ The first 'nmem' variables are the cross-sectional areas.
                Insert the current row of the statics matrix as the coefficient
                vector. 
            """
            a[(j+1)*nmem:(j+2)*nmem] = b            
            lp.add(LinearConstraint(a,pi,con_type="=", name="EQ-{0:.0g}-LC{1:.0g}".format(i,lc)))
    
    """ Constraints: stress constraints 
        These are expressed in terms of member forces, i.e.
        -sigma_lb*A[i] <= N[i] <= sigma_ub*A[i]
    """
    for j, lc in enumerate(truss.load_cases):
        for i, mem in truss.members.items():
            a = np.zeros(lp.nvars())
            a[i] = -sub[i]
            a[(j+1)*nmem+i] = 1
            lp.add(LinearConstraint(a,0,con_type="<", name="STR-UB-{0:.0g}-LC{1:.0g}".format(i,lc)))
            
            a[i] = -slb[i]
            a[(j+1)*nmem+i] = -1
            lp.add(LinearConstraint(a,0,con_type="<", name="STR-LB-{0:.0g}-LC{1:.0g}".format(i,lc)))
    
    return lp

def truss_milp(truss,profiles,ulb=-1e3,uub=1e3):
    """ Discrete sizing optimization of trusses using MILP formulation 
        
        It is assumed that all members have the same available set of profiles and
        each member has its own profile, i.e. no profile grouping.
    
        input:
            profiles .. list of available profiles. Profiles are RHS, SHS, IPE, etc. objects.
            ulb, uub .. bounds for nodal displacements
    """
    
    """ Do unit scaling. By default, the system is (N,mm). Perhaps
        (kN,cm) would be better    
    """
    Nscale = 1e-3
    uscale = 1e-1
    Escale = Nscale/(uscale**2)
    Ascale = uscale**2
    
    ulb *= uscale
    uub *= uscale
    
    """ Create variables """
    dvars = []
    c = []
    
    nmem = len(truss.members)   # number of members
    nprof = len(profiles)       # number of profiles
    ndof = truss.f.nfree_dofs   # number of degrees of freedom
    nlc = len(truss.load_cases) # number of load cases
    
    """ Statics matrix """
    B = truss.f.statics_matrix()  
    
    """ Calculate cross-sectional resistance of members and round down.
        By rounding, we try to avoid numerical issues in the computations.
    """
    NRd = [np.floor(p.NRd)*Nscale for p in profiles]
    
    """ Buckling resistance of members. This must be calculated for all profiles """
    NbRd = np.zeros([nprof,nmem])
    
    """ Bounds for force variables """
    Nlb = np.zeros([nprof,nmem])
    Nub = np.zeros([nprof,nmem])

    """ Design variables: member profile selection
        The assumption here is that all members can take all profiles from the list.
        If member grouping is used or if different profile alternatives are used for
        different members, this needs to be changed.
    """
    
    for i, mem in truss.members.items():
        L = mem.length*uscale
        k0 = mem.E/L*Escale
        nlb = np.sum(np.maximum(B[:,i],0)*ulb) + np.sum(np.minimum(B[:,i],0)*uub)
        nub = np.sum(np.minimum(B[:,i],0)*ulb) + np.sum(np.maximum(B[:,i],0)*uub)
        for j, p in enumerate(profiles):                        
            dvars.append(BinaryVariable("y({0:.0f},{1:.0f})".format(i,j),
                                            section=p,
                                            target={"objects":[mem],"property":"profile"}))
            """ Calculate buckling resistance of the member for profile 'p' """
            mem.profile = p
            NbRd[j,i] = np.floor(mem.NbRd[0])*Nscale
            Nlb[j,i] = k0*p.A*nlb*Ascale
            Nub[j,i] = k0*p.A*nub*Ascale
            c.append(p.weight()/uscale*L)
    
  
    """ Design variables: member forces 
        Force variables are N[i,j,k], where i is member, j is profile and k is load case
    
    """
    for lc in truss.load_cases:
        for i, mem in truss.members.items():            
            for j in range(nprof):
                dvars.append(Variable(name="N({0:.0f},{1:.0f},{2:.0f})".format(i,j,lc),lb=Nlb[j,i],ub=Nub[j,i],
                                          target=None,value=1))
                c.append(0)
    
    """ Design variables: nodal displacements 
        u[j,k]
    """
    for k in range(nlc):
        for j in range(ndof):
            dvars.append(Variable(name="u({0:.0f},{1:.0f})".format(j,k),
                                      lb=ulb,ub=uub,target=None,value=0.0))
            c.append(0)
            
    """ Initialize optimization problem with the created variables """
    lp = OptimizationProblem(name="Truss LP",variables=dvars,structure=truss)
    
    """ Objective function: truss weight """
    obj = LinearObjective("Weight", c)
    lp.add(obj)
    
    """ Constraints: unique profile selection """
    Aeq = np.zeros([nmem,lp.nvars()])
    for i, var in enumerate(lp.vars):
        """ Find the profile selection variables based on their name """
        s = var.name
        if s[0] != 'y':
            break
        else:
            """ The first number in the name of the variable corresponds to the member index """
            m = re.findall(r'\d+',s)
            Aeq[int(m[0]),i] = 1
            
    for i, a in enumerate(Aeq):
        lp.add(LinearConstraint(a,1,con_type="=", name="Sum y({0:.0f},j)=1".format(i)))
    
    """ Constraints: equilibrium equations """
    B0 = np.zeros([ndof,lp.nvars()])
    
    for i in range(nmem):
        B0[:,(nmem*nprof + i*nprof):(nmem*nprof + (i+1)*nprof)] += np.reshape(B[:,i],(ndof,1))
        
    
    for j, lc in enumerate(truss.load_cases):
        """ Load vector for the load case, including only unsupported dofs """
        p = np.delete(truss.f.global_load_vector(lc),truss.f.supp_dofs)*Nscale
        """ a is a placeholder for the coefficient vector of a linear constraint """
        a = np.zeros(lp.nvars())
        for i, (b,pi) in enumerate(zip(B0,p)):
            """ The first 'nmem' variables are the cross-sectional areas.
                Insert the current row of the statics matrix as the coefficient
                vector. 
            """
            lp.add(LinearConstraint(b,pi,con_type="=", name="EQ-{0:.0g}-LC{1:.0g}".format(i,lc)))
    
    """ Constraints: compatibility conditions """
    nx = lp.nvars()
    Nndx = nmem*nprof
    for k, lc in enumerate(truss.load_cases):
        for i, mem in truss.members.items():            
            k0 = mem.E/mem.length*Escale/uscale
            for j, p in enumerate(profiles):
                a = np.zeros(nx)
                """ Upper bound """
                a[i*nprof+j] = Nub[j,i]              # y variable
                a[Nndx] = -1  # N variable
                ustart = nmem*nprof + nlc*nmem*nprof + k*ndof
                uend = ustart +ndof
                a[ustart:uend] = k0*p.A*B[:,i]*Ascale                  # u variable
                lp.add(LinearConstraint(a,Nub[j,i],con_type="<", name="COMP-UB({0:.0f},{1:.0f},{2:.0f})".format(i,j,k)))
    
                """ Lower bound """
                a[Nndx] = 1  # N variable
                a[i*nprof+j] = -Nlb[j,i]             # y variable
                a[ustart:uend] *= -1    # u variable
                lp.add(LinearConstraint(a,-Nlb[j,i],con_type="<", name="COMP-LB({0:.0f},{1:.0f},{2:.0f})".format(i,j,k)))
                
                Nndx +=1
    """ Constraints: stress constraints 
        These are expressed in terms of member forces, i.e.
        -NRd[j]*y[i,j] <= N[i,j,k] <= NRd[j]*y[i,j]
    """
    Nndx = nmem*nprof
    for k, lc in enumerate(truss.load_cases):
        for i, mem in truss.members.items():            
            for j, Nrd in enumerate(NRd):
                a = np.zeros(lp.nvars())
                # Tension resistance 
                a[i*nprof+j] = -Nrd # -Nub[j,i] #               # y variable                
                a[Nndx] = 1  # N variable
                lp.add(LinearConstraint(a,0,con_type="<", name="STR-UB({0:.0f},{1:.0f},{2:.0f})".format(i,j,k)))
            
                #a[i] = -slb[i]
                # Compression resistance, including buckling 
                a[Nndx] = -1
                a[i*nprof+j] = -min(Nrd,NbRd[j,i])# Nlb[j,i] # 
                lp.add(LinearConstraint(a,0,con_type="<", name="STR-LB({0:.0f},{1:.0f},{2:.0f})".format(i,j,k)))
    
                Nndx += 1
    
    return lp

if __name__ == '__main__':
    #t = three_bar_truss(L=3000,F1=-200e3,F2=-250e3)
    t = ten_bar_truss(L=3000,F=200e3)
    
    #lp = truss_lp(t,0,1e8)
    #LPsolver = LP()
    #sol = LPsolver.solve(lp,verb=True)
    
    P = []
    
    hmax = 110
    hmin = 50
    tmin = 3.0
    
    for key, val in shs_profiles.items():
        if val['h'] <= hmax and val['h'] >= hmin and val['t'] >= tmin:
            P.append(SHS(val['h'],val['t']))
    
    lp = truss_milp(t,P,ulb=-200,uub=200)
    
    MILPsolver = MILP()
    sol = MILPsolver.solve(lp,verb=True)
    
    
    


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela


class PointLoad:

    def __init__(self, coord, F, load_id=0, scale=1):

        self.load_id = load_id
        self.node = None
        self._coord = coord
        self.F = F
        self.scale = scale

    @property
    def Fx(self):
        return self.F[0]

    @Fx.setter
    def Fx(self, val):
        self.F[0] = val

    @property
    def Fy(self):
        return self.F[1]

    @Fy.setter
    def Fy(self, val):
        self.F[1] = val

    @property
    def Fz(self):
        return self.F[2]

    @Fz.setter
    def Fz(self, val):
        self.F[2] = val

    @property
    def coord(self):
        return self.node.coord

    @coord.setter
    def coord(self, val):
        # 2D
        if len(val) == 2:
            x, y = val
            self.node.x = x
            self.node.y = y
        # 3D
        elif len(val) == 3:
            x, y, z = val
            self.node.x = x
            self.node.y = y
            self.node.z = z


if __name__ == '__main__':
    pl = PointLoad([0, 0], [0, 100, 0])
    print(pl)


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
    Cost optimization of end plate connections
    
    @author: Kristo Mela
    @date: 31.1.2020
"""

import math
import numpy as np
from copy import deepcopy
from openpyxl import Workbook, load_workbook
import matplotlib.pyplot as plt
import sys

   
import metku.optimization.structopt as sopt
from metku.structures.steel.plates import THICKNESSES
from metku.eurocodes.en1993.en1993_1_8.en1993_1_8 import SHEAR_ROW, ROW_OUTSIDE_BEAM_TENSION_FLANGE
from metku.eurocodes.en1993.en1993_1_8.en1993_1_8 import bolt_sizes

class EndPlateOpt(sopt.OptimizationProblem):
    """
    Class cost optimization of end plate connections

    """

    def __init__(self,
                 name="End-plate Optimization Problem",
                 structure=None,                 
                 constraints={'geometry':True},
                 variables={'bp':False, 'etop': False},
                 plate_thicknesses=THICKNESSES,
                 obj=None,
                 target_stiffness=None,
                 Sdev = 0.1):
        """ Constructor
            input:
                structure .. EndPlateJoint class object
                constraints .. list of constraints
                plate_thicknesses .. list of available plate thicknesses
                
        """
        super().__init__(name=name,
                         structure=structure)        
        self.plate_thicknesses = plate_thicknesses
        self.stiffness_dev = Sdev
        self.target_stiffness = target_stiffness
        self.Srigid = 0.0
        
        if variables is None:
            variables={'bp':False, 'etop': False}
            
        self.create_variables(variables)
        self._variables = variables
        
        
        if constraints is None:
            constraints = {
                'geometry': True,
                'bending_resistance': True,
                'shear_resistance': True,
                'target_stiffness': True,
                'rigid_joint': False,
                'pinned_joint': False,
                'initial_stiffness':False
            }
        
        
        
        if target_stiffness is None:
            constraints["target_stiffness"] = False
                        
        self.constraints = constraints
        self.create_constraints(**self.constraints)

        if obj is not None:
            self.create_objective(obj)
        elif obj is None:
            self.create_objective('cost')

            

    @property
    def available_constraints(self):
        """
        Returns all available constraints and their corresponding keyword
        :return:
        """
        return dict(geometry=False,
                    bending_resistance=False,
                    shear_resistance=False,
                    target_stiffness=False,
                    rigid_joint=False,
                    pinned_joint=False,
                    initial_stiffness=False)

    def vertical_row_distance_constraints(self, i):
        """
        Creates a constraint that states that two adjacent
        bolt rows need to be at least 2.2*d0 apart from each other
        vertically.
        :param i: index of the row
        :return: constraint functions in a dict
        """

        con_funs = {}
        d0 = self.structure.bolt_rows[0].bolt.d0
        """ Condition: z[i+1]-z[i] >= 2.2*d0 """
        rhs = 2.2*d0
        a = np.zeros(self.nvars())
        
        for j, var in enumerate(self.vars):
            if var.name == ('z' + str(i)):
                a[j] = 1
            elif var.name == ('z' + str(i+1)):
                a[j] = -1

        def p(x):
            self.substitute_variables(x)
            return self.structure.bolt_rows[i].z - self.structure.bolt_rows[i+1].z - 2.2*d0
            #return self.structure.bolt_rows[i+1].z_top - self.structure.bolt_rows[i].z_top - 2.2*d0

        #con_funs = {f'Row {i+2} - Row{i+1} >= 2.2*d0': p}
              
        con_name = f'Row {i+1} - Row{i} >= 2.2*d0'
        #con_fun = p        
        #return con_name, con_fun
        return sopt.LinearConstraint(a,rhs,'>',name=con_name)
    
    def extension_edge_distance(self):
        """ Limits the location of the row above the top flange of the beam
            etop + 0.5*hb - z1 >= 1.2*d0 or
            etop - z1 >= 1.2*d0 - 0.5*hb
        """
    
        a = np.zeros(self.nvars())
    
        for i, var in enumerate(self.vars):
            if var.name == 'eTop':
                a[i] = 1
            elif var.name == 'z1':
                a[i] = -1
                
    
        d0 = self.structure.bolt_rows[0].bolt.d0
        rhs = 1.2*d0 - 0.5*self.structure.beam.h
    
        return sopt.LinearConstraint(a,rhs,'>',name='Extension edge distance')
    
    def plate_edge_distance(self):
        """ Limits the location of the rows from the edge of end plate
            0.5*bp - x >= 1.2*d0
            x + 0.5*dw <= 0.5*bp
            ->
            0.5*bp - x >= max(1.2*d0,0.5*bp)
        """
    
        a = np.zeros(self.nvars())
    
        for i, var in enumerate(self.vars):
            if var.name == 'bp':
                a[i] = 0.5
            elif var.name == 'x':
                a[i] = -1
                
    
        d0 = self.structure.bolt_rows[0].bolt.d0
        dw = self.structure.bolt_rows[0].bolt.dw
        rhs = max(1.2*d0,0.5*dw)
    
        return sopt.LinearConstraint(a,rhs,'>',name='End plate edge distance')
    
    
    def moment_resistance_constraints(self):
        """ Moment resistance constraint:
            MjEd <= MjRd(x)
        """
        def Mcon(x):
            self.substitute_variables(x)
            #return 1-self.structure.MjRd*1e-6/self.structure.MjEd
            #return self.structure.MjEd-self.structure.MjRd*1e-6
            return self.structure.MjEd/(self.structure.MjRd*1e-6)-1
        
        con_name = "1-MjRd/MjEd <= 0"
        con_fun = Mcon
        
        return con_name, con_fun

    def shear_resistance_constraints(self):
        """ Moment resitance constraint:
            VjEd <= VjRd(x)
        """
        def Vcon(x):
            self.substitute_variables(x)
            #return 1-self.structure.VjRd*1e-3/self.structure.VjEd
            return self.structure.VjEd-self.structure.VjRd*1e-3
            #return self.structure.VjEd/(self.structure.VjRd*1e-3)-1
        
        con_name = "1-VjRd/VjEd <= 0"
        con_fun = Vcon
        
        return con_name, con_fun
    
    def stiffness_constraint(self,initial=False):
        """ Requirement for target stiffness
            input:
                S_target .. target stiffness (kNm)
                p .. allowed deviation from target stiffness (percent)
                
                Sj(x) <= (1+p)*S_target
                Sj(x) >= (1-p)*S_target
        """

        print(initial)

        def Scon_ub(x):
            """ Sj(x) <= (1+p)*S_target """
            self.substitute_variables(x)
            #return self.structure.rotational_stiffness()*1e-6/((1+self.stiffness_dev*1e-2)*self.target_stiffness) - 1
            if initial:
                return 1e-2*(self.structure.Sj_ini()*1e-6-(1+self.stiffness_dev*1e-2)*self.target_stiffness)
            else:
                return 1e-2*(self.structure.rotational_stiffness()*1e-6-(1+self.stiffness_dev*1e-2)*self.target_stiffness)

        def Scon_lb(x):
            """ Sj(x) >= (1-p)*S_target """
            self.substitute_variables(x)
            if initial:
                return 1e-2*(self.structure.Sj_ini()*1e-6-(1-self.stiffness_dev*1e-2)*self.target_stiffness)
            else:
            #return self.structure.rotational_stiffness()*1e-6/((1-self.stiffness_dev*1e-2)*self.target_stiffness) - 1
                return 1e-2*(self.structure.rotational_stiffness()*1e-6-(1-self.stiffness_dev*1e-2)*self.target_stiffness)
        
        cons = {}
        cons["Stiffness upper bound"] = Scon_ub
        cons["Stiffness lower bound"] = Scon_lb
        
        return cons
    
    def rigid_joint_constraint(self):
        """ Requirement that threshold initial rotational stiffness is exceeded
            so that the joint can be regarded as rigid
            input:
                S_rigid .. limit stiffness (kNm)
                
                S_j_init(x) >= S_rigid
                
        """

        def Scon_rigid(x):
            """ Sj_ini(x) >= S_rigid """
            self.substitute_variables(x)
            #return self.structure.rotational_stiffness()*1e-6/((1-self.stiffness_dev*1e-2)*self.target_stiffness) - 1
            return 1e0*(self.structure.Sj_ini()*1e-6-self.Srigid)
            #return self.structure.Sj_ini()*1e-6/self.Srigid-1
        
        con_name = "Stiffness for rigid joint"
        con_fun = Scon_rigid        
        
        return con_name, con_fun
    
    
    def create_constraints(self,
                           geometry=False,
                           bending_resistance=False,
                           shear_resistance=False,
                           target_stiffness=False,
                           rigid_joint=False,
                           pinned_joint=False,
                           initial_stiffness=False,
                           ):
        """
        Creates constraints

        NOTE! The name of these kwargs must match the key in the cons -dict
        created in constraint function creator method
        e.g.
        CORRECT:
        cons = {'compression': comp_fun, 'tension': ten_fun}
        FALSE:
        cons = {'comp_func': comp_fun, 'tension_con': ten_fun}

        :param geometry:        
        :return:
        """
        self.cons = []
        
        
        if geometry:
            for i in range(1,self.structure.nrows):
                if self.structure.bolt_rows[i].type != SHEAR_ROW:
                    con = self.vertical_row_distance_constraints(i)
                    self.add(con)
                """
                p_name, p_fun = self.vertical_row_distance_constraints(i)
                con = sopt.NonLinearConstraint(name=p_name,
                                               con_fun=p_fun,
                                               con_type='>',
                                               fea_required=False)                
                            #vars=mem)
                """
                    
                
            if self._variables['etop']:
                con = self.extension_edge_distance()
                self.add(con)
            
            if self._variables['bp']:
                con = self.plate_edge_distance()
                self.add(con)
            
        
        if bending_resistance:
            m_name, m_fun = self.moment_resistance_constraints()
            mcon = sopt.NonLinearConstraint(name=m_name,
                                            con_fun=m_fun,
                                            con_type='<',
                                            fea_required=False)
            self.add(mcon)

        if shear_resistance:
            v_name, v_fun = self.shear_resistance_constraints()
            mcon = sopt.NonLinearConstraint(name=v_name,
                                            con_fun=v_fun,
                                            con_type='<',
                                            fea_required=False)
            self.add(mcon)
            
        if target_stiffness:
            Scons = self.stiffness_constraint(initial_stiffness)
            
            for key, value in Scons.items():
                if key.find('upper') != -1:
                    """ upper bound """
                    ctype = '<'
                else:
                    """ lower bound """
                    ctype = '>'
                self.add(sopt.NonLinearConstraint(name=key,
                                                     con_fun=value,
                                                     con_type=ctype,
                                                     fea_required=False)
                         )

        if rigid_joint:
            r_name, r_fun = self.rigid_joint_constraint()
            
            self.add(sopt.NonLinearConstraint(name=r_name,
                                              con_fun=r_fun,
                                              con_type='>',
                                              fea_required=False)
                     )
            

    def create_variables(self,variables={'bp':False, 'etop': False}):
        """
        Creates design variables
        
        By default, the design variables are:
            tp .. end plate thickness
            x .. horizontal distance of all bolts from center axis of the beam
            zi .. vertical distance of bolt row i from top edge of the plate
            bp .. width of the end plate
            
        Alternatively, the zi could be replaced by:
            z1 .. distance of the top row from the top of the plate
            pi .. distance of adjacent rows
            -> this could be set to all rows such that the distance between
               selected rows would be equal. If zi variables are used,
               equality constraints need to be added to control the distance
               between rows.
        
        """
        self.vars = []

        """ End plate thickness """
        tp_var = sopt.Variable("tp",
                               lb=min(THICKNESSES),
                               ub=max(THICKNESSES),
                               target={"property":"t","objects":[self.structure.end_plate]},
                               value=self.structure.tp)
        
        self.add(tp_var)
        
        """ Distance from bolt rows from axis of the beam """
        d0 = self.structure.bolt_rows[0].bolt.d0
        dw = self.structure.bolt_rows[0].bolt.dw
        hb = self.structure.beam.h
        bb = self.structure.beam.b
        tb = self.structure.beam.tw
        tbf = self.structure.beam.tf
        bc = self.structure.col.b
        tc = self.structure.col.tw
        rc = self.structure.col.r
        af = self.structure.weld_f
        aw = self.structure.weld_w
        x_min = max(0.5*tb + math.sqrt(2)*aw + 0.5*dw,
                    0.5*tc + rc + 0.5*dw,
                    1.2*d0)
        # Distance from column flange edge
        x_max = min(0.5*bc-1.2*d0,
                    0.5*bc-0.5*dw)
        
        axis_x_var = sopt.Variable("x",
                                   lb=x_min,
                                   ub=x_max,
                                   target={"property":"xbolts","objects":[self.structure]},
                                   value=0.5*self.structure.bp-self.structure.ebolts)
        
        self.add(axis_x_var)
        
        """ Vertical distance of bolt rows 
        """
        if variables['etop']:
            """ If the top extension of the end plate is a variable,
                take the extension limit equal to 5*d0.
            """
            etop_ub = 5*d0
            lb = math.ceil(0.5*dw + math.sqrt(2)*af + 1.2*d0)
            etop_var = sopt.Variable("eTop",lb=lb, ub=etop_ub, target={"property":'etop','objects':[self.structure]},
                                  value=self.structure.etop)
            self.add(etop_var)
        else:
            """ If the top extension is not a variable, take the initial
                extension as the limit
            """
            etop_ub = self.structure.etop
            lup = self.structure.etop
        
        if variables['bp']:
            """ Create variable for end plate width """
            #bp_lb = bb + math.ceil(2*math.sqrt(2)*af)
            bp_lb = bb
            bp_ub = bc + 100.0
            bp_var = sopt.Variable("bp",lb=bp_lb, ub=bp_ub, target={"property":'bp','objects':[self.structure]},
                                  value=bb)
        
            self.add(bp_var)
        # zlb = max(1.2*d0,0.5*dw)
        # zub = hb
        # lb = zlb
        # Bounds for the first row, above the top flange of the beam
        zlb = math.ceil(0.5*hb + math.sqrt(2)*af + 0.5*dw)
        zub = 0.5*hb + etop_ub
        #zub = hb
        #lb = zlb
        self.add(sopt.Variable("z1",lb=zlb, ub=zub, target={"property":'z','objects':[self.structure.bolt_rows[0]]},
                                  value=self.structure.bolt_rows[0].z))
        
        zub = math.floor(0.5*hb - tbf - math.sqrt(2)*af - 0.5*dw)
        zlb = -zub
        
        for i, row in enumerate(self.structure.bolt_rows):
            if row.type is not SHEAR_ROW and row.loc_end_plate is not ROW_OUTSIDE_BEAM_TENSION_FLANGE:            
                self.add(sopt.Variable("z"+f"{i+1}",lb=zlb, ub=zub, target={"property":'z','objects':[row]},
                                  value=row.z))                
                zub -= 2.2*d0
        
        """
        for i, row in enumerate(self.structure.bolt_rows):
            if row.type is not SHEAR_ROW and loc_end_plate is not ROW_OUTSIDE_BEAM_TENSION_FLANGE:
                if i == 0:
                    "" First row above tension flange (assume extended end plate
                        connection.)
                    ""
                    ub = min(zub,math.floor(lup-math.sqrt(2)*af-0.5*dw))
                    #lb = zlb
                elif i == 1:
                    ub = zub
                    lb = max(zlb,math.ceil(lup+math.sqrt(2)*af+0.5*dw+tb))
                else:
                    lb += 2.2*d0
                    
                self.add(sopt.Variable("z"+f"{i+1}",lb=lb, ub=ub, target={"property":'z_top','objects':[row]},
                                  value=0.5*hb+self.structure.etop-row.z))
        """
        
    def create_objective(self,name='cost'):

        def obj_fun_cost(x):
            self.substitute_variables(x)            
            return self.structure.cost()
        
        def obj_fun_stiffness(x):
            self.substitute_variables(x)            
            return self.structure.Sj_ini()*1e-6
        
        def obj_fun_moment(x):
            self.substitute_variables(x)            
            return self.structure.bending_resistance()*1e-6
        
        if name == 'cost':
            fun = obj_fun_cost
            sense = 'MIN'
        elif name == 'stiffness':
            fun = obj_fun_stiffness
            sense = 'MAX'
        elif name == 'moment':
            fun = obj_fun_moment
            sense = 'MAX'

        obj = sopt.ObjectiveFunction(name=name.capitalize(),
                                     obj_fun=fun,
                                     obj_type=sense,
                                     fea_required=False)
        self.add(obj)


def optimize_for_target_stiffness(conn,Starget,rho=1.0,initial=False):
    
            
    print(initial)
    # Create problem
    problem = EndPlateOpt(name="Minimum cost, target stiffness",
                          structure=conn,                                
                          constraints={
                                    'geometry': True,
                                    'bending_resistance': True ,
                                    'shear_resistance': True,
                                    'target_stiffness': True,
                                    'initial_stiffness': initial,
                                },
                          variables={'bp':True, 'etop': True},
                          target_stiffness=Starget,
                          Sdev=rho)        
    
    return problem

def optimize_for_stiffness(conn,problem_type='MAX'):
                
    # Create problem
    if problem_type == "MAX":
        pname= "Stiffness maximization"        
    else:
        pname= "Stiffness minimization"        
        
    
    problem = EndPlateOpt(name="Stiffness maximization",
                          structure=conn,                          
                          constraints={
                                    'geometry': True,
                                    'bending_resistance': False ,
                                    'shear_resistance': False,
                                    'target_stiffness': False,
                                    'initial_stiffness': False,
                                },
                          obj='stiffness',                     
                          variables={'bp':True, 'etop': True})    

    if problem_type == 'MIN':
        problem.obj.obj_type = "MIN"
    
    return problem

def optimize_for_max_moment_resistance(conn):
                
    # Create problem
    problem = EndPlateOpt(name="Moment resistance maximization",
                          structure=conn,                          
                          constraints={
                                    'geometry': True,
                                    'bending_resistance': False ,
                                    'shear_resistance': False,
                                    'target_stiffness': False,
                                    'initial_stiffness': False,
                                },
                          obj='moment',                     
                          variables={'bp':True, 'etop': True})        
    
    return problem

def optimize_for_rigid_joint(conn,Srigid):

    problem = EndPlateOpt(name="Minimum cost, rigid joint",
                          structure=conn,                                
                          constraints={
                                    'geometry': True,
                                    'bending_resistance': True ,
                                    'shear_resistance': True,
                                    'rigid_joint': True,
                                    'target_stiffness': False,
                                    'initial_stiffness': True,
                                },
                          variables={'bp':True, 'etop': True},
                          target_stiffness=Srigid,
                          Sdev=0.1)
    
    problem.Srigid = Srigid
    
    return problem

def write_excel(problem,x_best,f_best,filename="Results.xlsx"):
    
    def print_rows():
        for row in sheet.iter_rows(values_only=True):
            print(row)
    
    #wb = Workbook()
    try:
        wb = load_workbook(filename=filename)
        sheet = wb.active
    except:
        wb = Workbook()
        sheet = wb.active
        
        # Create column names
        col_names = ['Rows','Bolt size', problem.obj.name]
        col_names += [var.name for var in problem.vars]
        if problem.constraints['target_stiffness']:
            if problem.constraints['initial_stiffness']:
                col_names += ['Sj_ini']
            else:
                col_names += ['Sj']
        else:
            col_names += ['Sj,ini']
        col_names += ['MjRd']
        col_names += ['VjRd']
        sheet.append(col_names)        
    
    var_vals = [x for x in x_best]
    problem(x_best)
    col_vals = [problem.structure.nrows,problem.structure.bolt_size,
                f_best]
    col_vals += var_vals
    if problem.constraints['target_stiffness']:
        if problem.constraints['initial_stiffness']:
            Sj = problem.structure.Sj_ini()*1e-6
        else:
            Sj = problem.structure.rotational_stiffness()*1e-6
        col_vals += [Sj,
                     problem.structure.MjRd*1e-6,
                     problem.structure.VjRd*1e-3]        
    else:
        col_vals += [problem.structure.Sj_ini()*1e-6,
                     problem.structure.MjRd*1e-6,
                     problem.structure.VjRd*1e-3]
    
    #sheet.append(var_names)
    
    sheet.append(col_vals)
    #print_rows()
    
    wb.save(filename=filename)
    
    
    



def run_optimization_series(conn,Starget,problem_type='rigid_joint',
                            initial=False):
    """ Runs a series of optimization problems for the joint 'conn'
        Varied parameters:
            i) bolt size
            ii) bolt grade
            iii) number of bolt rows
    
    """
    
    res = []    
    
    res_dir = 'P:\\INTRA_HOME\\00_Tutkimus\\IXConnections\\'
    #filename= res_dir + 'Diaz_Joint_Results.xlsx'
    
    filename= res_dir + problem_type.capitalize() + '_NEW2' + '_Diaz.xlsx'
    logfile = res_dir + problem_type.capitalize() + '_NEW2_opt.log'
    #filename= res_dir + 'Max_Stiffness_Results.xlsx'
    #fig_name_base = 'RigidJoint_M'
    fig_dir = 'P:\\INTRA_HOME\\00_Tutkimus\\IXConnections\\figures\\'
    #fig_name_base = 'DiazJoint_M'
    #fig_name_base = 'Max_Stiffness_M'
    fig_name_base = problem_type.capitalize() + '_NEW2' + '_Diaz_M'
    
    
    for d in reversed([12, 16, 20, 24]): #bolt_sizes:
        #sys.stdout = open(logfile, "a")
        print(f"**** COST OPTIMIZATION FOR {conn.nrows:2.0f} ROWS AND BOLT DIAMETER {d:2.0f}")
        
        conn.bolt_size = d
        if problem_type == 'rigid_joint':
            problem = optimize_for_rigid_joint(conn, Starget)
        elif problem_type == 'target_stiffness':
            problem = optimize_for_target_stiffness(conn, Starget, initial=initial)
        elif problem_type == 'max_stiffness':
            problem = optimize_for_stiffness(conn,'MAX')
        elif problem_type == 'min_stiffness':
            problem = optimize_for_stiffness(conn,'MIN')
        elif problem_type == 'max_moment':
            problem = optimize_for_max_moment_resistance(conn)
        else:
            print("Warning: unrecognized problem type!")
        
        
        solver = TrustRegionConstr()
        x0 = [var.value for var in problem.vars]
        
        if problem_type == "min_stiffness":
            x0[0] = problem.vars[0].lb
        
        print("Initial design:\n")
        for x, var in zip(x0,problem.vars):
            print(f'{var.name} = {x:4.2f}')
        
       

        f_best, x_best = solver.solve(problem, maxiter=500, x0=x0)

        #sys.stdout.close()
        
        #print("Tänne päästiin")
        #f_best, x_best, nit = solver.solve(problem, maxiter=500, x0=x0)
        #problem.num_iters = nit
        
        #problem(x_best, prec=5)
        
        fig_name = fig_name_base + str(d) + '_' + str(conn.nrows) + '_Rows.svg'
        #conn.draw(fig_dir + fig_name)
        #plt.close()
        res.append([d,problem,f_best,x_best])
        # Write results to Excel file
        #write_excel(problem,x_best,f_best,filename)

    # Finally, write initial parameters to Excel
    """
    wb = load_workbook(filename)
    sheet = wb.active
    
    sheet.insert_rows(idx=1,amount=6)
    if problem_type == 'rigid_joint':
        sheet["A1"] = 'Rigid joint cost minimization'
    else:
        sheet["A1"] = 'Cost minimization for target stiffness'
    sheet["A2"] = 'SjRigid'
    sheet["B2"] = problem.target_stiffness
    sheet["C2"] = 'kNm/rad'
    sheet["A3"] = 'MjEd'
    sheet["B3"] = problem.structure.MjEd
    sheet["C3"] = 'kNm'
    sheet["A4"] = 'VjEd'
    sheet["B4"] = problem.structure.VjEd
    sheet["C4"] = 'kN'
    sheet["A5"] = 'Column'
    sheet["B5"] = problem.structure.col.__repr__() + ' ' + problem.structure.col.material.__repr__()
    sheet["A6"] = 'Beam'
    sheet["B6"] = problem.structure.beam.__repr__() + ' ' + problem.structure.beam.material.__repr__()   

    wb.save(filename=filename)
    """

    return res

if __name__ == '__main__':
    
    from eurocodes.en1993.en1993_1_8 import en1993_1_8

    import structures.steel.end_plate_joint as ep
    from optimization.solvers.trust_region import TrustRegionConstr
    #from structures.steel.end_plate_joint import example_1
    
    #conn = ep.example_1()
    #nrows = range(2,5)
    nrows = [3]
    
    for r in reversed(nrows):
        conn = ep.diax_ex_rows(MjEd=40.0,VjEd=32.0,rows=r)
        #conn.MjEd = 0.5*round(0.8*conn.beam.MRd[0]*1e-6)
        #conn.MjEd = 180
        #conn = ep.example_diaz(connection='C')
        Starget = 16000.0
        Lbeam =  6000
        #Lbeam = 1200041.80200022598146
        Kb = conn.beam.E*conn.beam.Iy/Lbeam
        kb = 8
        Srigid = math.ceil(kb*Kb*1e-6)
        Starget = Srigid
        #Starget = 2.0*Kb*1e-6
        print("Target stiffness: {0:4.2f} kNm/rad".format(Starget))
    
        #conn.bending_resistance(True)
        #conn.rotational_stiffness(True)
        #conn.MjEd = 200.0 # 22.0
        #conn.VjEd = 200.0 # 26.325
        
        #conn.tp = 20
        res = run_optimization_series(conn,Starget,initial=False, problem_type='rigid_joint')
        #res = run_optimization_series(conn,Starget,initial=False, problem_type='max_stiffness')
    
    #problem = optimize_for_rigid_joint(conn, Srigid)
    
    """
    problem = EndPlateOpt(name="TEST",
                          structure=conn,                                
                          constraints={
                                    'geometry': True,
                                    'bending_resistance': True ,
                                    'shear_resistance': True,
                                    'stiffness': True
                                },
                          variables={'bp':True, 'etop': True},
                          target_stiffness=46000.0,
                          Sdev=0.1)
    """
    
    """
    solver = TrustRegionConstr()
    x0 = [var.value for var in problem.vars]
    f_best, x_best, nit = solver.solve(problem, maxiter=400, x0=x0)
    problem.num_iters = nit
    problem(x_best, prec=5)
    """
    # solver = GA(pop_size=50, mut_rate=0.15)
    # x0 = [1 for var in problem.vars]
    # fopt, xopt = solver.solve(problem, x0=x0, maxiter=10, plot=True)
    # problem(xopt)
    # frame.plot_deflection(100)



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
@author Viktor Haimi
"""

from enum import Enum


class LoadIDs(Enum):
    SLS_Characteristic = 1
    SLS_Frequent = 2
    SLS_Quasi_permanent = 3
    ULS = 4
    ACC = 5

